$SD.on("connected", (jsonObj) => connected(jsonObj));

function connected(jsn) {
  // Subscribe to the willAppear and other events
  $SD.on("uno.overlays.action.willAppear", (jsonObj) =>
    action.onWillAppear(jsonObj)
  );
  $SD.on("uno.overlays.action.keyUp", (jsonObj) => action.onKeyUp(jsonObj));
  $SD.on("uno.overlays.action.sendToPlugin", (jsonObj) =>
    action.onSendToPlugin(jsonObj)
  );
}

const action = {
  onWillAppear: function (jsn) {
    console.log(`[onWillAppear] ${JSON.stringify(jsn)}`);
  },

  onWillDisappear: function (jsn) {
    console.log(`[onWillDisappear] ${JSON.stringify(jsn)}`);
  },

  onKeyUp: function (jsn) {
    console.log(`[onKeyUp] ${JSON.stringify(jsn)}`);
    if (!jsn.payload.settings || !jsn.payload.settings["app-token"]) {
      $SD.api.showAlert(jsn.context);
      return;
    }
    this.sendCommandToApp(jsn.payload.settings, jsn.context);
  },

  onSendToPlugin: function (jsn) {
    console.log(`[onSendToPlugin] ${JSON.stringify(jsn)}`);
    if (jsn.payload) {
      $SD.api.setSettings(jsn.context, jsn.payload);
    }
  },

  sendCommandToApp: function (settings, context) {
    console.log(`[sendCommandToApp]: options = ${JSON.stringify(settings)}`);

    let payload = {};

    let command = JSON.parse(settings["app-command"]);

    if (command) {
      if (command.arguments && command.arguments.length) {
        // this is a special case for the ChangeOverlayField an ChangeCustomizationField command
        if (command.command === "ChangeOverlayField") {
          payload.id = settings["app-overlay"] || "";
          payload.fieldId = settings["app-overlay-field"] || "";

          let value = settings["app-overlay-field-operation-value"] || "";

          switch (settings["app-overlay-field-operation"]) {
            case "set":
              payload.command = "SetOverlayContentField";
              payload.value = value;
              break;
            case "numberIncrement":
              payload.command = "IncrementOverlayContentField";
              payload.value = value;
              break;
            case "numberDecrement":
              payload.command = "DecrementOverlayContentField";
              payload.value = value;
              break;
            case "checkboxToggle":
              payload.command = "ToggleOverlayContentField";
              break;
            case "timecontrolStart":
              payload.command = "ExecuteOverlayContentField";
              payload.value = "start";
              break;
            case "timecontrolPause":
              payload.command = "ExecuteOverlayContentField";
              payload.value = "pause";
              break;
            case "timecontrolReset":
              payload.command = "ExecuteOverlayContentField";
              payload.value = "reset";
              break;
            case "timecontrolPlay":
              payload.command = "ExecuteOverlayContentField";
              payload.value = "play";
              break;
            case "buttonClick":
              payload.command = "ExecuteOverlayContentField";
              payload.value = "execute";
              break;
          }
        } else if (command.command === "ChangeCustomizationField") {
          payload.fieldId = settings["app-overlay-field"] || "";

          let value = settings["app-overlay-field-operation-value"] || "";

          switch (settings["app-overlay-field-operation"]) {
            case "set":
              payload.command = "SetCustomizationField";
              payload.value = value;
              break;
            case "numberIncrement":
              payload.command = "IncrementCustomizationField";
              payload.value = value;
              break;
            case "numberDecrement":
              payload.command = "DecrementCustomizationField";
              payload.value = value;
              break;
            case "checkboxToggle":
              payload.command = "ToggleCustomizationField";
              break;
            case "timecontrolStart":
              payload.command = "ExecuteCustomizationField";
              payload.value = "start";
              break;
            case "timecontrolPause":
              payload.command = "ExecuteCustomizationField";
              payload.value = "pause";
              break;
            case "timecontrolReset":
              payload.command = "ExecuteCustomizationField";
              payload.value = "reset";
              break;
            case "timecontrolPlay":
              payload.command = "ExecuteCustomizationField";
              payload.value = "play";
              break;
            case "buttonClick":
              payload.command = "ExecuteCustomizationField";
              payload.value = "execute";
              break;
          }
        } else {
          payload.command = command.command;
          command.arguments.forEach((item) => {
            switch (item.id) {
              case "value":
                payload.value = settings["app-payload"] || "";
                break;
              case "id":
                payload.id = settings["app-overlay"] || "";
                break;
              case "content":
                {
                  let content = "";
                  try {
                    content = JSON.parse(settings["app-payload"]);
                  } catch (error) {
                    $SD.api.showAlert($SD.api.showAlert(context));
                    return;
                  }
                  payload.content = content;
                }
                break;
              default:
                payload.value = settings["app-payload"] || "";
                break;
            }
          });
        }
      } else {
        // we only have a command
        payload.command = command.command;
      }
    }

    const options = {
      method: "PUT",
      headers: {
        "content-type": "application/json",
      },
      body: JSON.stringify(payload),
    };

    // the url to send the command to
    const url =
      "https://app.overlays.uno/apiv2/controlapps/" +
      settings["app-token"] +
      "/api";
    // use this for development
    // const url = "http://localhost:4000/api";

    fetch(url, options)
      .then((response) => {
        if (!response.ok) {
          throw new Error("HTTP error, status = " + response.status);
        }
        $SD.api.showOk(context);
      })
      .catch((error) => {
        $SD.api.showAlert($SD.api.showAlert(context));
      });
  },
};
