// state data
let appToken = "";

let appCommandList = [];
let appCommandObject = {};

let appDatastoreId = "";

let appOverlayList = [];
let appOverlayId = "";
let oldAppOverlayId = "";
let appCustomization = null;

let appOverlayFieldId = "";
let appOverlayFieldOperationId = "";

let appOverlayFieldOperationValueElement = null;
let appOverlayFieldOperationValue = null;

let appPayload = "";

// the UI element that matches the type of payload the current command has
let errorMessage = "";
let hasOverlaySelection = false;
let validateAppStatus = "ok";
let fetchCustomizationStatus = "busy";
let fetchOverlayListStatus = "busy";
let fetchOverlayPayloadStatus = "ok";
let payloadUiElement;
let appName = "";
let appThumbnail = "";

const BaseUrl = "https://app.overlays.uno/apiv2/controlapps/";

$SD.on("connected", (jsn) => {
  if (jsn.hasOwnProperty("actionInfo")) {
    // get the current values of the plugin
    let settings = Utils.getProp(jsn, "actionInfo.payload.settings", {});

    // check if settings is a valid object containing data
    if (
      settings &&
      Object.keys(settings).length !== 0 &&
      settings.constructor === Object
    ) {
      // get the current values and remember them globally
      appToken = settings["app-token"] || "";
      appPayload = settings["app-payload"] || "";
      appOverlayId = settings["app-overlay"] || "";
      if (appOverlayId && appPayload) {
        oldAppOverlayId = appOverlayId;
      }
      appOverlayFieldId = settings["app-overlay-field"] || "";
      appOverlayFieldOperationId = settings["app-overlay-field-operation"];
      appOverlayFieldOperationValue =
        settings["app-overlay-field-operation-value"];

      appDatastoreId = settings["app-datastore"] || "";

      if (settings && settings["app-command"]) {
        try {
          appCommandObject = JSON.parse(settings["app-command"]);
        } catch (error) {}
      }

      // this will setup the payload UI
      rebuildUI();

      // if we have a token then fetch the commands
      if (appToken) {
        validateApp();
        fetchAppCustomization();
        fetchAppOverlays();
        fetchAppOverlayPayload();
      }
      return;
    }
  }

  // show the info
  showElement("message-info");
});

const save = () => {
  if ($SD) {
    var payload = {};
    payload["app-token"] = appToken;
    try {
      payload["app-command"] = JSON.stringify(appCommandObject);
    } catch (error) {
      payload["app-command"] = "";
    }
    payload["app-overlay"] = appOverlayId;
    if (payloadUiElement) {
      payload["app-payload"] = appPayload;
    } else {
      payload["app-payload"] = "";
    }
    payload["app-overlay-field"] = appOverlayFieldId;
    payload["app-overlay-field-operation"] = appOverlayFieldOperationId;
    payload["app-overlay-field-operation-value"] =
      appOverlayFieldOperationValue;

    payload["app-datastore"] = appDatastoreId;

    $SD.api.sendToPlugin($SD.uuid, $SD.actionInfo["action"], payload);
  }
};

const getAppJson = (event) => {
  // we have to press enter or tab
  let checkToken = false;
  if (!event) {
    checkToken = true;
  } else {
    const keyPressed = event.which || event.keyCode;
    if (keyPressed == 13 || keyPressed == 9) {
      checkToken = true;
    }
  }

  if (checkToken) {
    let token = document.getElementById("app-token").value;
    if (appToken != token) {
      appToken = token;
      validateApp();
      save();
    }
  }
};

const refreshContent = () => {
  // this will force the fetch
  oldAppOverlayId = undefined;
  fetchAppOverlayPayload();
};

// this is a debug function to initiate the REST API call directly from the stream deck plugin authoring enviroment without a stream deck device
const sendCommand = () => {
  var settings = {};
  settings["app-token"] = appToken;
  try {
    settings["app-command"] = JSON.stringify(appCommandObject);
  } catch (error) {
    settings["app-command"] = "";
  }
  settings["app-overlay"] = appOverlayId;
  if (payloadUiElement) {
    settings["app-payload"] = appPayload;
  } else {
    settings["app-payload"] = "";
  }
  settings["app-overlay-field"] = appOverlayFieldId;
  settings["app-overlay-field-operation"] = appOverlayFieldOperationId;
  settings["app-overlay-field-operation-value"] = appOverlayFieldOperationValue;
  settings["app-datastore"] = appDatastoreId;

  const url = BaseUrl + settings["app-token"] + "/api";
  // const url = "http://localhost:4000/api";

  let payload = {};

  let command = JSON.parse(settings["app-command"]);
  payload.command = command.command;
  if (command && command.arguments && command.arguments.length) {
    command.arguments.forEach((item) => {
      switch (item.id) {
        case "value":
          payload.value = settings["app-payload"] || "";
          break;
        case "id":
          payload.id = settings["app-overlay"] || "";
          break;
        case "content":
          {
            let content = "";
            try {
              content = JSON.parse(settings["app-payload"]);
            } catch (error) {
              console.log(error);
            }
            payload.content = content;
          }
          break;
        default:
          payload.value = settings["app-payload"] || "";
          break;
      }
    });
  }

  const options = {
    method: "PUT",
    headers: {
      "content-type": "application/json",
    },
    body: JSON.stringify(payload),
  };

  fetch(url, options)
    .then((response) => {
      if (!response.ok) {
        throw new Error("HTTP error, status = " + response.status);
      }
    })
    .catch((error) => {
      console.log(error);
    });
};

// check if the app token is valid
const validateApp = () => {
  validateAppStatus = "busy";
  appName = "";
  appThumbnail = "";
  appDatastoreId = "";
  rebuildUI();

  const errorHandler = () => {
    validateAppStatus = "error";
    errorMessage = "Token is invalid";
    appName = "";
    appThumbnail = "";
    appDatastoreId = "";
  };

  // new url will be
  const url = BaseUrl + `${appToken}`;

  fetch(url, {
    method: "GET",
    redirect: "follow",
  })
    .then((response) => {
      if (!response.ok) {
        errorHandler();
        rebuildUI();
        return;
      }
      return response.json();
    })
    .then((json) => {
      appName = json.name || "";
      appThumbnail = json.thumbnail || "";
      appDatastoreId = json.datastoreId || "";

      // this API call gives me the thumbnail without https
      if (appThumbnail && appThumbnail.indexOf("https://") === -1) {
        appThumbnail = "https:" + appThumbnail;
      }
      validateAppStatus = "ok";
      errorMessage = "";
      fetchAppApi();
      rebuildUI();
      // we need to save here to store the appDatastoreId
      // save();

      // send the analytics event to say that we opened stream deck
      fetch(
        `https://app.singular.live/apiv2/controlapps/${appToken}/analytics`,
        {
          method: "POST",
          headers: { "Content-Type": "application/json" },
          body: JSON.stringify({
            event: "streamdeckopened",
          }),
        }
      );
    })
    .catch((error) => {
      errorHandler();
      rebuildUI();
      return;
    });
};

const fetchAppApi = () => {
  const errorHandler = () => {
    validateAppStatus = "error";
    errorMessage = "This Uno overlay doesn't have an API defined!";
    appCommandList = [];
  };

  const url = BaseUrl + `${appToken}/api/json`;

  fetch(url, {
    method: "GET",
    redirect: "follow",
  })
    .then((response) => {
      if (!response.ok) {
        errorHandler();
        rebuildUI();
        throw new Error("HTTP error, status = " + response.status);
      }
      return response.json();
    })
    .then((json) => {
      appCommandList = json;

      // check if we have an overlay selection field.
      // If yes then we have to fetch the overlays from the API
      hasOverlaySelection = false;
      if (appCommandList && appCommandList.length) {
        appCommandList.forEach((command) => {
          if (command.arguments && command.arguments.length) {
            command.arguments.forEach((argument) => {
              if (argument.type === "overlaySelection") {
                hasOverlaySelection = true;
              }
            });
          }
        });
      }

      rebuildUI();
      fetchAppCustomization();
      fetchAppOverlays();
      return true;
    })
    .catch((error) => {
      errorHandler();
      rebuildUI();
      return false;
    });
};

const fetchAppCustomization = () => {
  if (appCustomization) {
    fetchCustomizationStatus = "ok";
    rebuildUI();
    return;
  }

  const url = BaseUrl + `${appToken}/api`;
  // const url = "http://localhost:4000/api";
  const options = {
    method: "PUT",
    headers: { "Content-Type": "application/json" },
    body: JSON.stringify({
      command: "GetCustomizationModel",
    }),
  };

  fetchCustomizationStatus = "busy";

  fetch(url, options)
    .then((response) => {
      if (!response.ok) {
        appCustomization = null;
        rebuildUI();
        throw new Error("HTTP error, status = " + response.status);
      }
      return response.json();
    })
    .then((json) => {
      appCustomization = json.payload;
      fetchCustomizationStatus = "ok";
      rebuildUI();
      fetchAppOverlayPayload();
    })
    .catch((error) => {
      fetchCustomizationStatus = "error";
      appCustomization = null;
      rebuildUI();
    });
};

const fetchAppOverlays = () => {
  if (hasOverlaySelection === false) {
    appOverlayList = [];
    fetchOverlayListStatus = "ok";
    return;
  }

  const url = BaseUrl + `${appToken}/api`;
  // const url = "http://localhost:4000/api";
  const options = {
    method: "PUT",
    headers: { "Content-Type": "application/json" },
    body: JSON.stringify({
      command: "GetOverlayModels",
    }),
  };

  fetchOverlayListStatus = "busy";

  fetch(url, options)
    .then((response) => {
      if (!response.ok) {
        appOverlayList = [];
        rebuildUI();
        throw new Error("HTTP error, status = " + response.status);
      }
      return response.json();
    })
    .then((json) => {
      appOverlayList = json.payload;
      fetchOverlayListStatus = "ok";
      rebuildUI();
      fetchAppOverlayPayload();
    })
    .catch((error) => {
      fetchOverlayListStatus = "error";
      appOverlayList = [];
      rebuildUI();
    });
};

const fetchAppOverlayPayload = () => {
  if (appOverlayId == oldAppOverlayId || !appOverlayId) {
    return;
  }
  oldAppOverlayId = appOverlayId;

  const url = BaseUrl + `${appToken}/api`;
  // const url = "http://localhost:4000/api";

  const options = {
    method: "PUT",
    headers: { "Content-Type": "application/json" },
    body: JSON.stringify({
      command:
        appOverlayId === "root" ? "GetCustomization" : "GetOverlayContent",
      id: appOverlayId,
    }),
  };

  fetchOverlayPayloadStatus = "busy";
  rebuildUI();

  fetch(url, options)
    .then((response) => {
      if (!response.ok) {
        appPayload = "{}";
        rebuildUI();
        throw new Error("HTTP error, status = " + response.status);
      }
      return response.json();
    })
    .then((json) => {
      appPayload = JSON.stringify(json.payload, null, 2);
      save();
      fetchOverlayPayloadStatus = "ok";
      rebuildUI();
    })
    .catch((error) => {
      appPayload = "{}";
      fetchOverlayPayloadStatus = "error";
      rebuildUI();
    });
};

const changeAppCommand = () => {
  let cmd;
  try {
    cmd = JSON.parse(
      document.getElementById("app-command-select-element").value
    );
  } catch {}

  if (!cmd) {
    return;
  }
  appCommandObject = cmd;
  appPayload = "";

  rebuildUI();
  save();
};

const changeOverlay = () => {
  appOverlayId = document.getElementById("app-overlay-select-element").value;
  if (appCommandObject && appCommandObject.command === "SetOverlayContent") {
    fetchAppOverlayPayload();
  } else {
    appPayload = "";
  }
  rebuildUI();
  save();
};

const changeOverlayField = () => {
  appOverlayFieldId = document.getElementById(
    "app-payload-select-field-element"
  ).value;
  // we changed the field so reset the value
  appOverlayFieldOperationValue = null;
  rebuildUI();
  save();
};

const changeOverlayFieldOperation = () => {
  appOverlayFieldOperationId = document.getElementById(
    "app-payload-select-field-operation-element"
  ).value;
  // we changed the operation so reset the value
  // appOverlayFieldOperationValue = null;
  rebuildUI();
  save();
};

const changeOverlayFieldOperationValue = () => {
  if (appOverlayFieldOperationValueElement.type === "checkbox") {
    appOverlayFieldOperationValue =
      appOverlayFieldOperationValueElement.checked;
  } else {
    appOverlayFieldOperationValue = appOverlayFieldOperationValueElement.value;
  }
  save();
  rebuildUI();
};

const changeAppPayload = () => {
  let value = "";
  if (payloadUiElement) {
    value = document.getElementById(payloadUiElement).value;
  }
  appPayload = value;
  rebuildUI();
  save();
};

// this function will rebuild the UI based on the state variables
const rebuildUI = () => {
  // show an error if we have one
  if (errorMessage) {
    setElementContent("message-error-content", errorMessage);
    showElement("message-error");
  } else {
    hideElement("message-error");
  }

  removeAndAddClass("app-token-info-error", "", "hidden");
  removeAndAddClass("app-token-info-busy", "", "hidden");
  removeAndAddClass("app-token-info-ok", "", "hidden");
  if (validateAppStatus === "busy") {
    removeAndAddClass("app-token-info-busy", "hidden", "");
  } else {
    if (validateAppStatus === "ok") {
      removeAndAddClass("app-token-info-ok", "hidden", "");
    } else {
      removeAndAddClass("app-token-info-error", "hidden", "");
    }
  }

  // hide the field for now. We will show it later if needed
  hideElement("app-overlay-select-container");
  hideElement("app-overlay-select-element");

  hideElement("app-payload-select-field-container");
  hideElement("app-payload-select-field-element");
  hideElement("app-payload-select-field-message");

  hideElement("app-payload-select-field-operation-container");

  hideElement("app-payload-select-field-operation-value-container");
  hideElement("app-payload-select-field-operation-select-element");
  hideElement("app-payload-select-field-operation-text-element");
  hideElement("app-payload-select-field-operation-textarea-element");
  hideElement("app-payload-select-field-operation-color-element");
  hideElement("app-payload-select-field-operation-number-element");
  hideElement("app-payload-select-field-operation-checkbox-element");
  hideElement("app-payload-select-field-operation-checkbox-label");

  hideElement("app-payload-json-container");
  hideElement("app-payload-json");

  hideElement("app-payload-field-container");
  hideElement("app-payload-field-text");
  hideElement("app-payload-field-number");
  hideElement("app-payload-field-select");
  hideElement("app-payload-field-textarea");

  // validateAppStatus
  if (appToken && validateAppStatus === "ok") {
    setElementValue("app-token", appToken);

    removeAndAddClass("app-token", "error", "");

    // hide the info
    hideElement("message-info");

    setElementContent("app-token-info-box-name", appName);
    let el = document.getElementById("app-token-info-box-image");

    // replace fit-in/150x150 with fit-in/200x200
    el.src = appThumbnail.replace("fit-in/150x150", "fit-in/300x300");
  } else {
    setElementValue("app-token", appToken || "");
    if (validateAppStatus === "error") {
      removeAndAddClass("app-token", "", "error");
    }
    hideElement("app-command-select-container");
    showElement("message-info");
    return;
  }

  // handle the command list
  rebuildCommandList();

  // should we show value of the action?
  if (
    appCommandObject &&
    appCommandObject.arguments &&
    appCommandObject.arguments.length
  ) {
    // handle the special case of SetCustomizationContent and ChangeCustomizationField
    if (appCommandObject) {
      if (
        appCommandObject.command === "SetCustomizationContent" ||
        appCommandObject.command === "ChangeCustomizationField"
      ) {
        appOverlayId = "root";
        // hasOverlaySelection = false;
      }
    }

    let displayOverlaySelection = false;
    appCommandObject.arguments.forEach((property) => {
      // should we show the value field at the end?
      let showValue = false;
      // if we have an overlay selection then we need to check if we have an overlay selected
      if (displayOverlaySelection) {
        if (appOverlayId && fetchOverlayListStatus === "ok") {
          showValue = true;
        }
      } else {
        // if we don't have an overlay selection then we can show the value
        showValue = true;
      }

      let element;
      switch (property.type) {
        case "json":
        case "JSON":
          showElement("app-payload-json-container");
          if (fetchOverlayPayloadStatus === "busy") {
            showElement("app-payload-json");
            setElementValue("app-payload-json", "Fetching content JSON ...");
          } else {
            element = "app-payload-json";
          }
          break;
        case "overlaySelection":
          displayOverlaySelection = true;
          showElement("app-overlay-select-container");
          rebuildOverlayList();
          break;
        case "fieldSelection":
          rebuildOverlayFieldList();
          rebuildOverlayFieldOperationsList();
          rebuildOverlayFieldOperationsValue();
          break;
        case "string":
        case "textarea":
          showElement("app-payload-field-container");
          element = "app-payload-field-textarea";
          break;
        case "text":
          showElement("app-payload-field-container");
          element = "app-payload-field-text";
          break;
        case "number":
        case "counter":
        case "normalizednumber":
          showElement("app-payload-field-container");
          element = "app-payload-field-number";
          break;
        case "selection":
          showElement("app-payload-field-container");
          element = "app-payload-field-select";
          rebuildPayloadFieldSelection(property);
          break;
        default:
          showElement("app-payload-field-container");
          element = "app-payload-field-text";
          break;
      }
      if (element) {
        payloadUiElement = element;
        if (!appPayload) {
          appPayload = property.default;
        }
        if (!appPayload) {
          appPayload = "";
        }

        showElement(element);
        setElementValue(element, appPayload);

        // check if there is a valid JSON in the text field
        if (property.type == "JSON") {
          if (appPayload === "") {
            oldAppOverlayId = undefined;
            fetchAppOverlayPayload();
            return;
          }
          let isOK = false;
          try {
            JSON.parse(appPayload);
            isOK = true;
          } catch (error) {}
          if (isOK) {
            removeAndAddClass(element, "invalid", "valid");
          } else {
            removeAndAddClass(element, "valid", "invalid");
          }
        }
      }
    });
  }
};

const rebuildCommandList = () => {
  const elSelect = document.getElementById("app-command-select-element");
  if (!appCommandList || !appCommandList.length) {
    // set the current command so we can see it in the UI while we load the API data from the server
    let option = document.createElement("option");
    option.value = appCommandObject.command || "";
    if (appCommandObject.title) {
      option.text = appCommandObject.title;
    } else {
      if (validateAppStatus === "error") {
        option.text = "Cannot fetch API Info";
      } else {
        option.text = "Fetching API Info ...";
      }
    }

    // add the current value of the command and disable the drop down box.
    elSelect.add(option, null);
    elSelect.disabled = true;
  } else {
    elSelect.disabled = false;
    while (elSelect.firstChild) {
      elSelect.removeChild(elSelect.firstChild);
    }

    // populate the list and select the current command
    let selectedItem, optgroup, firstItem;
    appCommandList.forEach((item) => {
      if (item.group) {
        optgroup = document.createElement("optgroup");
        optgroup.label = item.group;
        elSelect.add(optgroup, null);
      } else {
        const option = document.createElement("option");
        option.value = JSON.stringify(item);
        option.text = item.title || item.command;
        if (optgroup) {
          optgroup.appendChild(option);
        } else {
          elSelect.add(option, null);
        }
        if (!firstItem) {
          firstItem = item;
        }

        // select the command that we use or, if no command was selected yet use the first
        if (item.command == appCommandObject.command) {
          selectedItem = item;
        }
      }
    });
    if (selectedItem) {
      elSelect.value = JSON.stringify(selectedItem);
      appCommandObject = selectedItem;
    } else {
      // this will make sure we skip the group header
      if (firstItem) {
        elSelect.value = JSON.stringify(firstItem);
        appCommandObject = firstItem;
      } else {
        elSelect.value = JSON.stringify(appCommandList[0]);
        appCommandObject = appCommandList[0];
      }
    }
  }

  showElement("app-command-select-container");
};

const rebuildOverlayList = () => {
  const elSelect = document.getElementById("app-overlay-select-element");
  elSelect.querySelectorAll("option").forEach((option) => option.remove());

  if (!hasOverlays()) {
    // set the current command so we can see it in the UI while we load the API data from the server
    let option = document.createElement("option");
    option.value = appOverlayId || "";
    if (fetchOverlayListStatus === "error") {
      option.text = "Failed to fetch overlays";
    } else {
      option.text = "Fetching Overlays ...";
    }

    // add the current value of the command and disable the drop down box.
    elSelect.add(option, null);
    elSelect.disabled = true;
  } else {
    elSelect.disabled = false;

    // check if our command needs a slot filter
    let filterSlots = false;
    if (appCommandObject && appCommandObject.command) {
      const c = appCommandObject.command;
      if (
        c === "TakeOverlayFirstSlot" ||
        c === "TakeOverlayLastSlot" ||
        c === "TakeOverlayNextSlot" ||
        c === "TakeOverlayPreviousSlot" ||
        c === "TakeOverlaySlotName" ||
        c === "TakeOverlaySlotNumber"
      ) {
        filterSlots = true;
      }
    }

    // filter the list if needed
    const filteredAppOverlayList = filterSlots
      ? appOverlayList.filter((item) => item.hasSlots)
      : appOverlayList;

    // if there are no overlays with slots available, add a dummy entry
    if (!filteredAppOverlayList.length) {
      let option = document.createElement("option");
      option.value = "";
      option.text = "No overlays with slots available";
      appOverlayId = "";
      elSelect.add(option, null);
    } else {
      // populate the list and select the current command
      let selectedId;
      filteredAppOverlayList.forEach((item) => {
        const option = document.createElement("option");
        option.value = item.id;
        option.text = item.name;
        elSelect.add(option, null);

        // select the overlay that we use or, if no overlay was selected yet use the first
        if (item.id == appOverlayId) {
          selectedId = item.id;
        }
      });
      if (selectedId) {
        elSelect.value = selectedId;
      } else {
        elSelect.value = filteredAppOverlayList[0].id;
        appOverlayId = filteredAppOverlayList[0].id;
      }
    }
  }
  showElement("app-overlay-select-element");
};

const rebuildOverlayFieldList = () => {
  if (!hasOverlays()) {
    return;
  } else {
    const elSelect = document.getElementById(
      "app-payload-select-field-element"
    );

    showElement("app-payload-select-field-container");

    // find the selected overlay
    let selectedOverlayModel = getSelectedOverlayModel();
    let selectedOverlayModelGroups = getSelectedOverlayModelGroups();

    // if we cannot find the selected overlay we return
    if (!selectedOverlayModel || !selectedOverlayModel.length) {
      showElement("app-payload-select-field-message");
      return;
    }

    // filter the fields of the model to only show the ones that are supported
    const supportedFields = selectedOverlayModel.filter((field) => {
      return (
        field.type === "text" ||
        field.type === "textarea" ||
        field.type === "number" ||
        field.type === "counter" ||
        field.type === "normalizednumber" ||
        field.type === "selection" ||
        field.type === "checkbox" ||
        field.type === "timecontrol" ||
        field.type === "image" ||
        field.type === "color" ||
        field.type === "gradient" ||
        field.type === "button"
      );
    });

    // if there are no supported fields then return
    if (!supportedFields || !supportedFields.length) {
      showElement("app-payload-select-field-message");
      return;
    }

    elSelect.disabled = false;
    elSelect.disabled = false;
    while (elSelect.firstChild) {
      elSelect.removeChild(elSelect.firstChild);
    }

    // sort the selected fields by index
    supportedFields.sort((a, b) => {
      return a.index - b.index;
    });

    const fieldTypeToName = (type) => {
      switch (type) {
        case "text":
          return "Text";
        case "textarea":
          return "Text Area";
        case "number":
          return "Number";
        case "counter":
          return "Counter";
        case "normalizednumber":
          return "Normalized Number";
        case "selection":
          return "Dropdown";
        case "checkbox":
          return "Checkbox";
        case "timecontrol":
          return "Time Control";
        case "image":
          return "Image";
        case "color":
        case "gradient":
          return "Color";
        case "button":
          return "Button";
        default:
          return "Unknown";
      }
    };

    // we have groups so list the elements in groups
    let selectedId;
    if (selectedOverlayModelGroups && selectedOverlayModelGroups.length) {
      // remember all field IDs
      const allFieldIds = supportedFields.map((field) => field.id);

      selectedOverlayModelGroups.forEach((group) => {
        if (!group.childIds || !group.childIds.length) {
          return;
        }

        const optgroup = document.createElement("optgroup");
        optgroup.label = group.title;
        elSelect.add(optgroup, null);

        group.childIds.forEach((item) => {
          const field = supportedFields.find((field) => field.id === item);

          if (field) {
            // remove the field from the list
            const index = allFieldIds.indexOf(field.id);
            if (index > -1) {
              allFieldIds.splice(index, 1);
            }

            const option = document.createElement("option");
            option.value = field.id;
            option.text = fieldTypeToName(field.type) + " - " + field.title;
            optgroup.appendChild(option);

            // select the field that we use
            if (field.id == appOverlayFieldId) {
              selectedId = field.id;
            }
          }
        });
      });

      // now create a group for the left over fields
      if (allFieldIds.length) {
        const optgroup = document.createElement("optgroup");
        optgroup.label = "Miscellaneous";
        elSelect.add(optgroup, null);

        allFieldIds.forEach((item) => {
          const field = supportedFields.find((field) => field.id === item);

          if (field) {
            const option = document.createElement("option");
            option.value = field.id;
            option.text = fieldTypeToName(field.type) + " - " + field.title;
            optgroup.appendChild(option);

            // select the field that we use
            if (field.id == appOverlayFieldId) {
              selectedId = field.id;
            }
          }
        });
      }
    } else {
      // populate the list and select the current command
      supportedFields.forEach((item) => {
        const option = document.createElement("option");
        option.value = item.id;
        option.text = fieldTypeToName(item.type) + " - " + item.title;
        elSelect.add(option, null);

        // select the field that we use
        if (item.id == appOverlayFieldId) {
          selectedId = item.id;
        }
      });
    }
    if (selectedId) {
      elSelect.value = selectedId;
    } else {
      elSelect.value = supportedFields[0].id;
      appOverlayFieldId = supportedFields[0].id;
      // we changed the field so reset the value
      appOverlayFieldOperationValue = null;
    }
    showElement("app-payload-select-field-element");
  }
};

const rebuildOverlayFieldOperationsList = () => {
  if (!hasOverlays()) {
    return;
  }
  const selectedField = getSelectedField();
  if (!selectedField) {
    return;
  }

  const elSelect = document.getElementById(
    "app-payload-select-field-operation-element"
  );
  showElement("app-payload-select-field-operation-container");

  let operations = [
    {
      id: "set",
      title: "Set",
    },
  ];
  if (selectedField.type === "button") {
    operations = [{ id: "buttonClick", title: "Click" }];
  }
  if (
    selectedField.type === "number" ||
    selectedField.type === "counter" ||
    selectedField.type === "normalizednumber"
  ) {
    operations.push({
      id: "numberIncrement",
      title: "Increment",
    });
    operations.push({
      id: "numberDecrement",
      title: "Decrement",
    });
  }
  if (selectedField.type === "checkbox") {
    operations.push({
      id: "checkboxToggle",
      title: "Toggle",
    });
  }
  if (selectedField.type === "timecontrol") {
    operations = [
      {
        id: "timecontrolStart",
        title: "Start",
      },
      {
        id: "timecontrolPause",
        title: "Pause",
      },
      {
        id: "timecontrolReset",
        title: "Reset",
      },
      {
        id: "timecontrolPlay",
        title: "Play",
      },
    ];
  }

  elSelect.disabled = false;

  elSelect.querySelectorAll("option").forEach((option) => option.remove());

  // populate the list and select the current command
  let selectedId;
  operations.forEach((item) => {
    const option = document.createElement("option");
    option.value = item.id;
    option.text = item.title;
    elSelect.add(option, null);

    // select the operation that we use
    if (item.id == appOverlayFieldOperationId) {
      selectedId = item.id;
    }
  });
  if (selectedId) {
    elSelect.value = selectedId;
  } else {
    elSelect.value = operations[0].id;
    appOverlayFieldOperationId = operations[0].id;
  }
};

const rebuildOverlayFieldOperationsValue = () => {
  if (!hasOverlays()) {
    return;
  }
  const selectedField = getSelectedField();
  if (!selectedField) {
    return;
  }

  const showContainerAndField = (fieldId) => {
    showElement("app-payload-select-field-operation-value-container");
    showElement(fieldId);
    appOverlayFieldOperationValueElement = document.getElementById(fieldId);

    // we have an undefined value so we set it to the default value
    let displayValue = selectedField.defaultValue;
    if (appOverlayFieldOperationValue === null) {
      appOverlayFieldOperationValue = displayValue;
    } else {
      displayValue = appOverlayFieldOperationValue;
    }

    // set the checked for checkbox and the value for the rest
    if (selectedField.type === "checkbox") {
      appOverlayFieldOperationValueElement.checked = displayValue;
    } else {
      // make sure that the color is a hex value
      if (selectedField.type === "color" || selectedField.type === "gradient") {
        // use the solid color of a gradient structure
        if (displayValue && displayValue.type === "solid") {
          displayValue = displayValue.solidColor;
        }
        appOverlayFieldOperationValueElement.value =
          tinycolor(displayValue).toHexString();
      } else {
        appOverlayFieldOperationValueElement.value = displayValue;
      }
    }
  };

  // show the value fields for the selected operation
  if (
    selectedField.type === "number" ||
    selectedField.type === "counter" ||
    selectedField.type === "normalizednumber"
  ) {
    // configure the number element
    const elNumber = document.getElementById(
      "app-payload-select-field-operation-number-element"
    );
    if (selectedField.min !== undefined) {
      elNumber.min = selectedField.min;
    }
    if (selectedField.max !== undefined) {
      elNumber.max = selectedField.max;
    }
    if (selectedField.step !== undefined) {
      elNumber.step = selectedField.step;
    } else {
      if (selectedField.type === "normalizednumber") {
        elNumber.step = 0.1;
      } else {
        elNumber.step = 1;
      }
    }
    showContainerAndField("app-payload-select-field-operation-number-element");
  } else if (selectedField.type === "checkbox") {
    // only show the checkbox if the operation is set
    if (appOverlayFieldOperationId === "set") {
      showElement("app-payload-select-field-operation-checkbox-label");
      showContainerAndField(
        "app-payload-select-field-operation-checkbox-element"
      );
    }
  } else if (selectedField.type === "timecontrol") {
    // nothing for timecontrol
  } else if (
    selectedField.type === "color" ||
    selectedField.type === "gradient"
  ) {
    showContainerAndField("app-payload-select-field-operation-color-element");
  } else if (selectedField.type === "selection") {
    if (
      selectedField.source === "url" &&
      selectedField.sourceUrl &&
      selectedField.fetchingDataStatus === undefined
    ) {
      // set a flag in the field that we are fetching the data
      selectedField.fetchingDataStatus = "busy";
      selectedField.selections = [];
      // check if the url starts with http and add it
      if (selectedField.sourceUrl.indexOf("//") === 0) {
        selectedField.sourceUrl = "https:" + selectedField.sourceUrl;
      }
      // fetch the data from the URL and put it into the selections field
      fetch(selectedField.sourceUrl)
        .then((response) => {
          if (!response.ok) {
            throw new Error("HTTP error, status = " + response.status);
          }
          return response.json();
        })
        .then((json) => {
          selectedField.fetchingDataStatus = "ok";
          selectedField.selections = json;
          rebuildOverlayFieldOperationsValue();
        })
        .catch((error) => {
          selectedField.fetchingDataStatus = "error";
          rebuildOverlayFieldOperationsValue();
        });
    }

    showContainerAndField("app-payload-select-field-operation-select-element");
    // configure the selection
    const elSelect = document.getElementById(
      "app-payload-select-field-operation-select-element"
    );
    elSelect.querySelectorAll("option").forEach((option) => option.remove());
    elSelect.disabled = false;
    if (selectedField.selections && selectedField.selections.length) {
      let selectedId;
      selectedField.selections.forEach((item) => {
        const option = document.createElement("option");
        option.value = item.id;
        option.text = item.title;
        elSelect.add(option, null);

        // select the overlay that we use or, if no overlay was selected yet use the first
        if (item.id == appOverlayFieldOperationValue) {
          selectedId = item.id;
        }
      });
      if (selectedId) {
        elSelect.value = selectedId;
      } else {
        elSelect.value = selectedField.selections[0].id;
        appOverlayFieldOperationValue = selectedField.selections[0].id;
      }
    } else {
      const option = document.createElement("option");
      option.value = "";
      if (selectedField.fetchingDataStatus === "busy") {
        option.text = "Fetching data ...";
      } else if (selectedField.fetchingDataStatus === "error") {
        option.text = "Failed to fetch data";
      } else {
        option.text = "No selection available";
      }
      elSelect.add(option, null);
      elSelect.disabled = true;
      appOverlayFieldOperationValue = null;
    }
  } else if (selectedField.type === "textarea") {
    showContainerAndField(
      "app-payload-select-field-operation-textarea-element"
    );
  } else if (
    selectedField.type === "color" ||
    selectedField.type === "gradient"
  ) {
    showContainerAndField("app-payload-select-field-operation-color-element");
  } else if (selectedField.type === "button") {
    // nothing for buttons
  } else {
    showContainerAndField("app-payload-select-field-operation-text-element");
  }
};

const rebuildPayloadFieldSelection = (property) => {
  const selections = property.selections;

  const elSelect = document.getElementById("app-payload-field-select");
  elSelect.querySelectorAll("option").forEach((option) => option.remove());

  if (!selections || !selections.length) {
    const option = document.createElement("option");
    option.value = "";
    option.text = "No selection available";
    elSelect.add(option, null);
    elSelect.disabled = true;
  } else {
    elSelect.disabled = false;

    let selectedId;
    selections.forEach((item) => {
      const option = document.createElement("option");
      option.value = item.id;
      option.text = item.title;
      elSelect.add(option, null);

      // select the overlay that we use or, if no overlay was selected yet use the first
      if (item.id == appPayload) {
        selectedId = item.id;
      }
    });
    if (selectedId) {
      elSelect.value = selectedId;
    } else {
      elSelect.value = selections[0].id;
      appPayload = selections[0].id;
    }
  }

  showElement("app-payload-field-select");
};

const setElementContent = (elementID, content) => {
  document.getElementById(elementID).textContent = content;
};
const setElementValue = (elementID, value) => {
  document.getElementById(elementID).value = value;
};
const showElement = (elementID) => {
  removeAndAddClass(elementID, "hidden", "shown");
};
const hideElement = (elementID) => {
  removeAndAddClass(elementID, "shown", "hidden");
};
const removeAndAddClass = (elementID, classRemove, classAdd) => {
  el = document.getElementById(elementID);
  if (!el) {
    console.log("Element not found: " + elementID);
    return;
  }

  if (classRemove) {
    el.classList.remove(classRemove);
  }
  if (classAdd) {
    el.classList.add(classAdd);
  }
};

const hasOverlays = () => {
  if (appOverlayList && appOverlayList.length) {
    return true;
  }
  if (appCommandObject) {
    if (
      appCommandObject.command === "SetCustomizationContent" ||
      appCommandObject.command === "ChangeCustomizationField"
    ) {
      return true;
    }
  }
  return false;
};

const getSelectedOverlayModel = () => {
  if (appCommandObject) {
    if (
      appCommandObject.command === "SetCustomizationContent" ||
      appCommandObject.command === "ChangeCustomizationField"
    ) {
      if (appCustomization) {
        return appCustomization.model;
      }
      return;
    }
  }

  const result = appOverlayList.find((item) => {
    return item.id == appOverlayId;
  });
  if (result) {
    return result.model;
  }
};

const getSelectedOverlayModelGroups = () => {
  if (appCommandObject) {
    if (
      appCommandObject.command === "SetCustomizationContent" ||
      appCommandObject.command === "ChangeCustomizationField"
    ) {
      if (appCustomization) {
        return appCustomization.groups;
      }
      return;
    }
  }

  const result = appOverlayList.find((item) => {
    return item.id == appOverlayId;
  });
  if (result) {
    return result.groups;
  }
};

const getSelectedField = () => {
  const selectedOverlayModel = getSelectedOverlayModel();
  if (!selectedOverlayModel) {
    return;
  }
  return selectedOverlayModel.find((item) => {
    return item.id == appOverlayFieldId;
  });
};
